/**
 * @file
 * @brief The mainline for the Netsweeper Squid URL Redirector
 *
 * This loads the configuration and starts the threads for each connection to each server.
 * A thread is created for each connection that is created.
 * These threads poll for requests from a common queue of requests.
 * Each thread is responsible for connection management of one connection.
 *
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>

#include "redir.h"
#include "protocol.h"
#include "config.h"
#include "utils.h"
#include "request.h"
#include "log.h"

/** @brief The default stack size for the thread we create is 2MB */
#define REDIR_STACK_SIZE (2 * 1024 * 1024)

/** @brief The number of FDs we need before adding one per server connection.
 *
 * This is roughly equal to 3 (stdin, stdout, stderr).  Add a few more for good
 * luck.
 */
#define BASE_FD_COUNT 5

/** @brief Total memory limit for server threads
 *
 * We will check the product of STACK_SIZE * thread count and make sure
 * it's smaller than this number.   This is so we dont just segfault
 * when we try to create too many threads
 **/
#define TOTAL_STACK_MEMORY_KB (2 * 1024 * 1024)

/** @brief Simple signal handler that sets the 'should_exit' flag */
void redir_signal_handler(int signo __attribute__ ((unused))) {
	redir_should_exit = 1;
}

/**
 * @brief The main function
 *
 * @param argc Number of arguments
 * @param argv String array of arguments
 * @return int 0 on success otherwise failure
 */
int main(int argc, char **argv)
{
	int i;
	Request *request;
	pthread_rwlock_t start_lock;
	Connection **connection_list = NULL;
	int servers_num = 0;
	struct rlimit rlim;
	int needed_fds;
	long long memory_requirements_kB;
	void *exit_data;

	// parse command line
	GlobalConf* conf = config_parse_command_line(argc, argv);
	if (conf == NULL) {
		exit(-2);
	}

	// Are we going to exceed our fd limit?
	// Number of FDs needed = BASEFDs + conns_per_serv * num_servers
	needed_fds = BASE_FD_COUNT + conf->ncons * conf->nsp_num;
	if ( getrlimit (RLIMIT_NOFILE, &rlim) != 0 ||
	     (int)rlim.rlim_cur < needed_fds ) {
		fprintf (stderr, "Use ulimit -n to increase your maximum FDs.\n"
		         "Your current limit is too low for the requested number of "
		         "connections * number of servers\n");
		config_free_conf(conf);
		exit(-3);
	}

	// Are we likely to exceed stack memory with all these
	// server threads?
	memory_requirements_kB = conf->ncons * conf->nsp_num * (REDIR_STACK_SIZE / 1024 );
	if ( TOTAL_STACK_MEMORY_KB < memory_requirements_kB) {
		fprintf (stderr, "Your configuration exceeds our memory limits!\n"
		         "Your configuration: est %llu kilobytes\n"
		         "Hard coded limit  :     %ld kilobytes\n",
		         memory_requirements_kB, (long)TOTAL_STACK_MEMORY_KB);
		exit (-4);
	}

	// we can init logging now
	log_init(conf);

	// catch signals to exit gracefully
	signal(SIGINT,redir_signal_handler);
	signal(SIGQUIT,redir_signal_handler);
	signal(SIGTERM,redir_signal_handler);

	// init request dispatcher (request queue )
	if (redir_dispatcher_init(conf) < 0) {
		exit(-1);
	}

	// prepare connection handlers for all servers from command line
	connection_list = config_init_connections(conf, &servers_num);

	if (servers_num == 0) {
		log_send(LOG_ERR, "No policy servers to connect\n");
	}

	// init thread start synchronization lock
	// (it prevents redirector from rejecting input request
	//  when threads are not started yet and there is no active connections)
	pthread_rwlock_init(&start_lock, NULL);

	// start server threads
	for (i = 0; i < servers_num ;i++) {
		pthread_attr_t attr;

		// pass the start lock to thread and add a lock
		// that thread should unlock when it is running
		connection_list[i]->start_lock = &start_lock;
		pthread_rwlock_rdlock(&start_lock);

		// increment the count of running threads
		redir_st_run();

		// limit the stack size since we create a thread per connection
		pthread_attr_init(&attr);
		pthread_attr_setstacksize(&attr, REDIR_STACK_SIZE);

		// Try to start the connection thread
		if (pthread_create(&connection_list[i]->thread_id, &attr, server_thread, (void*) connection_list[i]) != 0) {
			log_send(LOG_ERR, "Can not start server connection thread\n");
			// unlock the start lock because the thread will not unlock it
			pthread_rwlock_unlock(&start_lock);
			// decrement the count of running threads
			redir_st_exit();
		}
	}

	// Wait until all thread start (have connected or failed to connect)
	pthread_rwlock_wrlock(&start_lock);
	// we don't need this lock anymore
	pthread_rwlock_unlock(&start_lock);
	pthread_rwlock_destroy(&start_lock);

	// if no one thread is running we can't work and exit now
	if (redir_get_threads() == 0) {
		log_send(LOG_ERR, "No one server can be connected. Exit.");
		// exit(-1);
	}

	// Set non blocking mode on file descriptors for squid read and write
	if (fcntl(conf->input_fd, F_SETFL, O_NONBLOCK)) {
		log_send(LOG_ERR, "Can't set Non Block on input file descriptor (stdin)\n");
	}
	if (fcntl(conf->output_fd, F_SETFL, O_NONBLOCK)) {
		log_send(LOG_ERR, "Can't set Non Block on output file descriptor (stdout)\n");
	}

	// The main loop
	while (1) {

		// get request from stdin
		request = redir_get_request(conf);

		// send request to processing
		if(request != NULL) {
			redir_dispatch_request(request, 0);
			// check if the dispatcher queue is overloaded
			// (it can wait here if queue is full)
			redir_check_dispatch_space();
		}

		// check if redir should exit
		if (redir_should_exit == 1) {
			break;
		}
		// check if at least one thread is still running
		else if (redir_get_threads() == 0) {
			log_send(LOG_ERR, "No one server can be connected. Exit.");
			break;
		}

	}

	// Shutdown redirector
	// (the order is important we should be sure that queue is empty
	// before connection shutdown)

	// close input to prevent new request queueing
	close(conf->input_fd);

	// flush all waiting requests
	redir_do_flush_request_queue(1);

	// close the output
	close(conf->output_fd);

	// wake up all sleeping threads
	redir_dispatcher_broadcast();

	// wait all connection threads shutdown
	for (i = 0; i < servers_num ;i++) {
		pthread_join(connection_list[i]->thread_id, &exit_data);
	}

	// terminate dispatcher
	redir_dispatcher_terminate();

	// free used memory
	config_free_connections(connection_list, servers_num);
	config_free_conf(conf);
	exit(-1);
}
