/**
 * @file
 * @brief A simple select based socket server for unit tests
 *
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#ifndef __SOCKETSERVER_HEADER_H__
#define __SOCKETSERVER_HEADER_H__

/** @brief The recv buffer size */
#define CLIENT_BUFFER_SIZE 25000

// Forward declaration of the socket server
struct socketserver;

/** @brief Our client socket object, passed to the recv and connect callbacks */
struct client_socket
{
	/** @brief The socket fd associated with this connection */
	int socket;

	/** @brief Flag used to indicate we should disconnect this socket, call socketserver_connection_shutdown */
	int remove;

	/** @brief our receive buffer (null terminated) */
	char rxbuffer[CLIENT_BUFFER_SIZE];
	/** @brief The length of the receive buffer */
	int rxlen;

	/** @brief Our transmit buffer */
	char txbuffer[CLIENT_BUFFER_SIZE];
	/** @brief The size of our transmit buffer */
	int txlen;

	/** @brief The last time we had activity on this connection */
	int last_action_time;

	/** @brief Data pointer that can be utilized for this connection */
	void* tag;

	/** @brief Pointer to our server instance */
	struct socketserver* srv;
};

/** @brief Callback function used for connect/recv/disconnect */
typedef void (*SOCKETSERVER_CALLBACK)(struct client_socket*);

/** @brief Callback function used to free the socketserver tag */
typedef void (*TAGFREE_CALLBACK)(void *tag);

/** @brief Our socket server object, used for a single server instance */
struct socketserver {
	/** @brief The threadid of this server instance */
	pthread_t threadid;
	/** @brief Our start lock for this server */
	pthread_rwlock_t start_lock;
	/** @brief Our flag used to indicate the server should be running */
	int running;
	/** @brief Our flag used to force server to close all connections */
	int reset;
	/** @brief The port this server uses */
	int port;
	/** @brief The connect callback */
	SOCKETSERVER_CALLBACK connect_callback;
	/** @brief The recv data callback */
	SOCKETSERVER_CALLBACK recv_callback;
	/** @brief The disconnect callback */
	SOCKETSERVER_CALLBACK disconnect_callback;

	/** @brief Our tag free function, called during shutdown when set with socketserver_tagfree_func */
	TAGFREE_CALLBACK tagfree_func;

	/** @brief A private data pointer that can be utilized for this server */
	void* tag;
};


struct socketserver* socketserver_create(int port, SOCKETSERVER_CALLBACK connect_callback, SOCKETSERVER_CALLBACK recv_callback, SOCKETSERVER_CALLBACK disconnect_callback);

void socketserver_tagfree_func(struct socketserver* svr, TAGFREE_CALLBACK func);

int socketserver_send(struct client_socket* s, char* buffer, int len);
int socketserver_connection_shutdown(struct client_socket* s);
void socketserver_signal_shutdown(struct socketserver* svr);
void socketserver_waitfor_shutdown(struct socketserver* svr);

void socketserver_signal_reset_connections(struct socketserver* svr);
void socketserver_waitfor_reset_connections(struct socketserver* svr);

#endif
